#include "stratify.h"

#include "samplewindow.h"
#include "pngimage.h"

void 
SampleWindow::
SetOctree(SamplerOctree *_tree)
{
    tree = _tree;
}

void
SampleWindow::
SetMask(Array<char> *_mask)
{
    mask = _mask;
}

void
SampleWindow::
Load(void) {
    char name[10];
    printf("file: ");
    scanf("%s", name);
    printf("got '%s'\n", name);
    FILE *fp = fopen(name, "r");
    if (!fp) {
        errorf(("not found"));
        return;
    }
    cloud->normal.Resize(0);
    cloud->vertex.Resize(0);
    mask->Resize(0);
    double x, y, z;
    while (fscanf(fp, "%lg %lg %lg", &x, &y, &z) == 3) {
        cloud->vertex.PushBack(Vector3(x, y, z));
        cloud->normal.PushBack(Vector3(0, 0, 1));
        mask->PushBack(1);
    }
    fclose(fp);
}

SamplerOctree * 
SampleWindow::
GetOctree(void) const
{
    return tree;
}

static
void drawpoint(const Vector3 &p)
{
    glBegin(GL_POINTS);
        glVertex3fv(&p[0]); 
    glEnd();
}

static
void drawsphere(const Vector3 &p, double r)
{
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glTranslated(p[0], p[1], p[2]);
    glutSolidSphere(r, 16, 16);
    glPopMatrix();
}

static
void drawbox(const Vector3 &min, const Vector3 &max)
{
    glBegin(GL_LINE_LOOP);
    glVertex3d(min[0], min[1], min[2]);
    glVertex3d(max[0], min[1], min[2]);
    glVertex3d(max[0], max[1], min[2]);
    glVertex3d(min[0], max[1], min[2]);
    glEnd();
    glBegin(GL_LINE_LOOP);
    glVertex3d(min[0], min[1], max[2]);
    glVertex3d(max[0], min[1], max[2]);
    glVertex3d(max[0], max[1], max[2]);
    glVertex3d(min[0], max[1], max[2]);
    glEnd();
    glBegin(GL_LINES);
    glVertex3d(min[0], min[1], min[2]);
    glVertex3d(min[0], min[1], max[2]);
    glVertex3d(max[0], min[1], min[2]);
    glVertex3d(max[0], min[1], max[2]);
    glVertex3d(max[0], max[1], min[2]);
    glVertex3d(max[0], max[1], max[2]);
    glVertex3d(min[0], max[1], min[2]);
    glVertex3d(min[0], max[1], max[2]);
    glEnd();
}

static
void drawcube(const Vector3 &min, const Vector3 &max)
{
    glBegin(GL_QUADS);
    glNormal3d(0, 0, 1);
    glVertex3d(min[0], min[1], min[2]);
    glVertex3d(max[0], min[1], min[2]);
    glVertex3d(max[0], max[1], min[2]);
    glVertex3d(min[0], max[1], min[2]);
    glNormal3d(0, 0, -1);
    glVertex3d(min[0], min[1], max[2]);
    glVertex3d(max[0], min[1], max[2]);
    glVertex3d(max[0], max[1], max[2]);
    glVertex3d(min[0], max[1], max[2]);
    glNormal3d(0, 1, 0);
    glVertex3d(min[0], min[1], min[2]);
    glVertex3d(max[0], min[1], min[2]);
    glVertex3d(max[0], min[1], max[2]);
    glVertex3d(min[0], min[1], max[2]);
    glNormal3d(0, -1, 0);
    glVertex3d(min[0], max[1], min[2]);
    glVertex3d(max[0], max[1], min[2]);
    glVertex3d(max[0], max[1], max[2]);
    glVertex3d(min[0], max[1], max[2]);
    glNormal3d(1, 0, 0);
    glVertex3d(min[0], min[1], min[2]);
    glVertex3d(min[0], max[1], min[2]);
    glVertex3d(min[0], max[1], max[2]);
    glVertex3d(min[0], min[1], max[2]);
    glNormal3d(-1, 0, 0);
    glVertex3d(max[0], min[1], min[2]);
    glVertex3d(max[0], max[1], min[2]);
    glVertex3d(max[0], max[1], max[2]);
    glVertex3d(max[0], min[1], max[2]);
    glEnd();
}

static void drawtree(SamplerOctree *node, int level = 1)
{
    drawbox(node->box.GetMin(), node->box.GetMax());
}

static void drawhit(SamplerOctree *node, int level = 1)
{
    if (node->element.size() > 0) {
        glEnable(GL_POLYGON_OFFSET_FILL);
        glPolygonOffset(0, 3);
        glColor3d(1, 1, 1);
        drawcube(node->box.GetMin(), node->box.GetMax());
        glDisable(GL_POLYGON_OFFSET_FILL);
        glColor3d(0, 0, 0);
        drawbox(node->box.GetMin(), node->box.GetMax());
    }
}

static void grab(int w, int h)
{
    BufferUchar b;
    b.Resize(w, h, 3);
    GlutWMReadPixels(0, 0, w, h, GL_RGB, b.GetBuffer());
    PngImage i;
    i.Load(b);
    i.Store("grab.png");
}

SampleWindow::SampleWindow() 
{ 
    selected = -1;
    cloud = NULL; 
    drag = 0;
    color_list = 0;
    point_size = 3.0;
    screen_size = 512;
    multiply = 1;
    tree = NULL;
    dhit = false;
    dtree = false;
    dmesh = true;
    dblack = false;
    dcandidate = false;
    dremoved = false;
    doriginal = true;
    dxpensive = false;
    dnormals = false;
    dall = false;
    pointrender = false;
}

SampleWindow::~SampleWindow() 
{ 
    ; 
}

void 
SampleWindow::
SetSampleSize(double p, double s)
{
    point_size = p;
    screen_size = s;
}

double 
SampleWindow::
GetPointSize(void)
{
    int new_screen_size = MIN(GetWidth(), GetHeight());
    return GetScale() * multiply * point_size * new_screen_size / screen_size;
}

double 
SampleWindow::
GetSphereSize(void)
{
    return point_size * multiply * GetSize() / screen_size;
}

void 
SampleWindow::
SetCloud(Cloud *_cloud) 
{ 
    cloud = _cloud; 
}

Cloud *
SampleWindow::
GetCloud(void) const
{ 
    return cloud; 
}

void
SampleWindow::
UpdatePick(int x, int y)
{
    selected = PickByColor(x, y);
    if (selected > 0 && tree) {
        TriMesh *mesh = GetTriMesh();
        int f = st_dbg_from_face[selected];
        int i0 = mesh->faces[f][0];
        int i1 = mesh->faces[f][1];
        int i2 = mesh->faces[f][2];
        Vector3 n0(&mesh->normals[i0][0]);
        Vector3 n1(&mesh->normals[i1][0]);
        Vector3 n2(&mesh->normals[i2][0]);
        debugf(("%d -> %d: %d, %d, %d", selected, f, i0, i1, i2));
        Vector3 n = cloud->normal[selected];
        debugf(("n: (%g, %g, %g)", n[0], n[1], n[2]));
    }
    GlutWMPostRedisplay();
}

void 
SampleWindow::
Button(int button, int state, int x, int y)
{
    if (glutGetModifiers() & GLUT_ACTIVE_CTRL) {
        if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) {
            drag = 1;
            UpdatePick(x, y);
        } else drag = 0;
    } else TriMeshWindow::Button(button, state, x, y);
}

void
SampleWindow::
Motion(int x, int y)
{
    if (drag) UpdatePick(x, y);
    else TriMeshWindow::Motion(x, y);
}

void
SampleWindow::
Keyboard(unsigned char c, int x, int y)
{
    switch (c)  {
        case '>':
            multiply *= 2;
            PostRedisplay();
            break;
        case '<':
            multiply /= 2;
            PostRedisplay();
            break;
        case 'x':
            dxpensive = !dxpensive;
            PostRedisplay();
            break;
        case 'b':
            dblack = !dblack;
            PostRedisplay();
            break;
        case 'c':
            dcandidate = !dcandidate;
            PostRedisplay();
            break;
        case 'r':
            dremoved = !dremoved;
            PostRedisplay();
            break;
        case 't':
            dtree = !dtree;
            PostRedisplay();
            break;
        case 'h':
            dhit = !dhit;
            PostRedisplay();
            break;
        case 'g':
            grab(GetWidth(), GetHeight());
            break;
        case 'm':
            dmesh = !dmesh;
            PostRedisplay();
            break;
        case 'o':
            doriginal = !doriginal;
            PostRedisplay();
            break;
        case 'n':
            dnormals = !dnormals;
            PostRedisplay();
            break;
        case 'p':
            pointrender = !pointrender;
            PostRedisplay();
            break;
        case 'a':
            dall = !dall;
            PostRedisplay();
            break;
        case 'l':
            Load();
            break;
        case '?':
            break;
        default:
            TriMeshWindow::Keyboard(c, x, y);
            break;
    }
}

void
SampleWindow::
DrawNormals(void)
{
    if (!dnormals) return; 
    glLineWidth(2);
    glColor3f(1, 1, 0);
    glDisable(GL_LIGHTING);
    glBegin(GL_LINES);
    double size = GetTriMesh()->feature_size();
    for (int i = 0; i < cloud->vertex.GetSize(); i++) {
        if (mask->at(i)) {
            glVertex3fv(&cloud->vertex[i][0]);
            glVertex3fv(&(cloud->vertex[i] + size * cloud->normal[i])[0]);
        }
    }
    glEnd();
}

void
SampleWindow::
DrawModel(void)
{
    glClearColor(0, 0, 0, 0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    if (pointrender) {
        glEnable(GL_LIGHTING);
        glDisable(GL_BLEND);
        glDisable(GL_POINT_SMOOTH);
        glPointSize(1);
        TriMeshWindow::SetupUniforms();
        glBegin(GL_POINTS);
        for (int i = 0; i < cloud->vertex.GetSize(); i++) {
            if (mask->at(i)) {
                glNormal3fv(&cloud->normal[i][0]);
                glVertex3fv(&cloud->vertex[i][0]);
            }
        }
        glEnd();
    } else {
        if (dmesh) {
            if (dblack) glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
            TriMeshWindow::DrawModel();
            glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
        }
        DrawSelected();
        if (!dtree) DrawSamples();
        DrawOctree();
        DrawNormals();
    }
}

void
SampleWindow::
DrawOctree(void)
{
    // draw octree 
    if (!tree) return;
    if (dtree) {
        glColor3d(0.7, 0.5, 0.5);
        glLineWidth(GetPointSize()/2);
        glDisable(GL_LIGHTING);
        tree->PreOrder(drawtree);
    }
    if (dhit) {
        glEnable(GL_LIGHTING);
        tree->PreOrder(drawhit);
    }
}

void
SampleWindow::
DrawSample(const Vector3 &p)
{
    if (dxpensive) drawsphere(p, GetSphereSize());
    else drawpoint(p);
}

void
SampleWindow::
DrawNeighbors(void)
{
    if (!tree) return;
    SamplerOctree *n = st_dbg_tree_clone->Find(st_dbg_cache[selected].point);
    double d = Norm(n->box.GetSize());
    SamplerOctree *neighbor[27];
    n->GetNeighbors(neighbor);
    glLineWidth(3);
    for (int i = 0; i < 27; i++) {
        if (neighbor[i]) { 
            glDisable(GL_LIGHTING);
            glColor3d(0, 0, 0);
            drawtree(neighbor[i]);
            glEnable(GL_LIGHTING);
            for (int j = 0; j < neighbor[i]->element.GetSize(); j++) {
                int p = neighbor[i]->element[j].index;
                if (Norm(st_dbg_cache[p].point - 
                            st_dbg_cache[selected].point) < d)
                    glColor3d(1, 1, 0);
                else glColor3d(0, 1, 1);
                DrawSample(st_dbg_cache[p].point);
            }
        }
    }
}

void
SampleWindow::
DrawOriginal(void)
{
    if (!tree) return;
    // draw node from which point came 
    SamplerOctree *n = st_dbg_from_node[selected];
    TriMesh *m = GetTriMesh();
    glDisable(GL_LIGHTING);
    glLineWidth(1);
    glColor3d(1, 1, 1);
    drawtree(n);
    // draw all triangle that hit originating box
    glEnable(GL_LIGHTING);
    glColor3d(0, 0, 1);
    glEnable(GL_POLYGON_OFFSET_FILL);
    glPolygonOffset(0, -3);
    // draw triangle from which point came
    glBegin(GL_TRIANGLES);
        glColor3f(0, 1, 0);
        int f = st_dbg_from_face[selected];
        glNormal3fv(&m->normals[m->faces[f][0]][0]);
        glVertex3fv(&m->vertices[m->faces[f][0]][0]);
        glNormal3fv(&m->normals[m->faces[f][1]][0]);
        glVertex3fv(&m->vertices[m->faces[f][1]][0]);
        glNormal3fv(&m->normals[m->faces[f][2]][0]);
        glVertex3fv(&m->vertices[m->faces[f][2]][0]);
        glColor3f(0, 0, 1);
        // draw triangles that hit original octree node
        for (int i = 0; i < n->element.GetSize(); i++) {
            int index = n->element[i].index;
            glNormal3fv(&m->normals[m->faces[index][0]][0]);
            glVertex3fv(&m->vertices[m->faces[index][0]][0]);
            glNormal3fv(&m->normals[m->faces[index][1]][0]);
            glVertex3fv(&m->vertices[m->faces[index][1]][0]);
            glNormal3fv(&m->normals[m->faces[index][2]][0]);
            glVertex3fv(&m->vertices[m->faces[index][2]][0]);
        }
    glEnd();
    glDisable(GL_POLYGON_OFFSET_FILL);
}

void 
SampleWindow::
DrawSelected(void)
{
    if (selected >= cloud->vertex.GetSize())
        selected = -1;
    if (selected >= 0) {
        glColor3d(1, 0, 1);
        DrawSample(cloud->vertex[selected]);
        if (doriginal) DrawOriginal();
        else DrawNeighbors();
    }
}

void
SampleWindow::
DrawSamples(void)
{
    // draw all vertices
    if (!dxpensive) {
        glDisable(GL_LIGHTING);
        glPointSize(GetPointSize());
        glEnable(GL_POINT_SMOOTH);
        glEnable(GL_BLEND);
        glHint(GL_POINT_SMOOTH, GL_NICEST);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    } else glEnable(GL_LIGHTING);
    // draw removed samples
    if (dremoved) {
        glColor3d(1, 0, 0);
        for (int i = 0; i < cloud->vertex.GetSize(); i++)
            if (mask->at(i) == 0) DrawSample(cloud->vertex[i]);
    }
    // draw candidates for removal
    if (dcandidate) {
        glColor3d(0, 1, 0);
        for (int i = 0; i < cloud->vertex.GetSize(); i++) {
            if (mask->at(i) == 2) DrawSample(cloud->vertex[i]);
            if (mask->at(i) == 0) DrawSample(cloud->vertex[i]);
        }
    }
    glColor3f(0, 0, 1);
    // draw all points
    for (int i = 0; i < cloud->vertex.GetSize(); i++) 
        if (mask->at(i) || dall) DrawSample(cloud->vertex[i]);
}

void
SampleWindow::
Save(void)
{
    if (!cloud || !mask) return;
    FILE *fp = fopen("sample.pts", "w");
    if (!fp) return;
    int npoints = 0;
    for (int i = 0; i < cloud->vertex.GetSize(); i++)
        if (mask->at(i)) npoints++;
    fprintf(fp, "%d\n", npoints);
    for (int i = 0; i < cloud->vertex.GetSize(); i++)  {
        if (mask->at(i)) {
            Vector3 p = cloud->vertex[i];
            Vector3 n = cloud->normal[i];
            fprintf(fp, "%g %g %g %g %g %g\n", p[0], p[1], p[2],
                    n[0], n[1], n[2]);
        }
    }
    fclose(fp);
    debugf(("Saved 'sample.pts' with %d samples", npoints));
}

static void i2color(int i, GLubyte *color)
{
    color[0] = (GLubyte) (i & 0xff);
    i >>= 8;
    color[1] = (GLubyte) (i & 0xff);
    i >>= 8;
    color[2] = (GLubyte) (i & 0xff);
}

void
SampleWindow::
ColorSamples(void)
{
    glClearColor(1, 1, 1, 1);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glDisable(GL_LIGHTING);
    glDisable(GL_BLEND);
    glPointSize(5);
    glBegin(GL_POINTS);
    // loop through each vertex drawing it in a unique color
    for (size_t i = 0; i < cloud->vertex.size(); i++) {
        GLubyte color[3];
        i2color(i, color);
        glColor3ubv(color);
        glVertex3fv(&cloud->vertex[i][0]);
    }
    glEnd();
}

static int color2i(GLubyte *color)
{
    int i = color[2];
    i <<= 8;
    i |= color[1];
    i <<= 8;
    i |= color[0];
    return (i == 0xffffff) ? -1: i;
}

int 
SampleWindow::
PickByColor(int mx, int my)
{
    GLubyte color[3];
    ColorSamples();
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    GlutWMReadPixels(mx, my, 1, 1, GL_RGB, color);
    return color2i(color); 
}
