#include "glutwm.h"
#include "modelwindow.h"
#include "vector.h"
#include "bbox.h"
#include "array.h"
#include "pngimage.h"

class BPTSWindow: public ModelWindow {
public:
    void Fit(void) {
        Vector3 c;
        BBox3 b;
        for (int i = 0; i < data.GetSize(); i++) {
            c += data[i];
            b.Merge(data[i]);
        }
        c /= data.GetSize();
        SetSize(b.GetSize()[0]);
        SetCenter(c);
    }
    void DrawModel(void) {
        glClearColor(0, 0, 0, 0);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        glColor3d(1, 1, 1);
        glDisable(GL_LIGHTING);
        glVertexPointer(3, GL_FLOAT, 0, &data[0][0]);
        glEnableClientState(GL_VERTEX_ARRAY);
        glDrawArrays(GL_POINTS, 0, data.GetSize());
    }

    int Load(const char *name) {
        data.Resize(0);
        FILE *fin = fopen(name, "rb");
        if (!fin) return 0;
        while (1) {
            Vector3 f;
            if (fread(&f[0], sizeof(float), 3, fin) != 3) break;
            data.PushBack(f);
        }
        fclose(fin);
        return 1;
    }
private:
    Array<Vector3> data;
};

int main(int argc, char *argv[])
{
    GlutWMOpen(&argc, argv);
    GlutWMInitWindowSize(512, 512);
    if (argc < 2) {
        fprintf(stderr, "Usage\n  viewbpts <bpts>\n");
        return 1;
    }
    for (int i = 1; i < argc; i++) {
        BPTSWindow *bpts = new BPTSWindow;
        if (!bpts->Load(argv[i])) {
            debugf(("Error loading '%s'", argv[i]));
            delete bpts;
        } else { 
            bpts->Fit();
            GlutWMCreateWindow(argv[i], bpts);
        }
    }
    GlutWMMainLoop();
    return 0;
}
